
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#include "config.h"
#include "sysprof-enums.h"
#include "sysprof-callgraph.h"
#include "sysprof-recording.h"

#define C_ENUM(v) ((gint) v)
#define C_FLAGS(v) ((guint) v)

/* enumerations from "sysprof-callgraph.h" */

GType
sysprof_callgraph_category_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (SYSPROF_CALLGRAPH_CATEGORY_UNCATEGORIZED), "SYSPROF_CALLGRAPH_CATEGORY_UNCATEGORIZED", "uncategorized" },
        { C_ENUM (SYSPROF_CALLGRAPH_CATEGORY_PRESENTATION), "SYSPROF_CALLGRAPH_CATEGORY_PRESENTATION", "presentation" },
        { C_ENUM (SYSPROF_CALLGRAPH_CATEGORY_A11Y), "SYSPROF_CALLGRAPH_CATEGORY_A11Y", "a11y" },
        { C_ENUM (SYSPROF_CALLGRAPH_CATEGORY_ACTIONS), "SYSPROF_CALLGRAPH_CATEGORY_ACTIONS", "actions" },
        { C_ENUM (SYSPROF_CALLGRAPH_CATEGORY_CONTEXT_SWITCH), "SYSPROF_CALLGRAPH_CATEGORY_CONTEXT_SWITCH", "context-switch" },
        { C_ENUM (SYSPROF_CALLGRAPH_CATEGORY_COREDUMP), "SYSPROF_CALLGRAPH_CATEGORY_COREDUMP", "coredump" },
        { C_ENUM (SYSPROF_CALLGRAPH_CATEGORY_CSS), "SYSPROF_CALLGRAPH_CATEGORY_CSS", "css" },
        { C_ENUM (SYSPROF_CALLGRAPH_CATEGORY_GRAPHICS), "SYSPROF_CALLGRAPH_CATEGORY_GRAPHICS", "graphics" },
        { C_ENUM (SYSPROF_CALLGRAPH_CATEGORY_ICONS), "SYSPROF_CALLGRAPH_CATEGORY_ICONS", "icons" },
        { C_ENUM (SYSPROF_CALLGRAPH_CATEGORY_INPUT), "SYSPROF_CALLGRAPH_CATEGORY_INPUT", "input" },
        { C_ENUM (SYSPROF_CALLGRAPH_CATEGORY_IO), "SYSPROF_CALLGRAPH_CATEGORY_IO", "io" },
        { C_ENUM (SYSPROF_CALLGRAPH_CATEGORY_IPC), "SYSPROF_CALLGRAPH_CATEGORY_IPC", "ipc" },
        { C_ENUM (SYSPROF_CALLGRAPH_CATEGORY_JAVASCRIPT), "SYSPROF_CALLGRAPH_CATEGORY_JAVASCRIPT", "javascript" },
        { C_ENUM (SYSPROF_CALLGRAPH_CATEGORY_KERNEL), "SYSPROF_CALLGRAPH_CATEGORY_KERNEL", "kernel" },
        { C_ENUM (SYSPROF_CALLGRAPH_CATEGORY_LAYOUT), "SYSPROF_CALLGRAPH_CATEGORY_LAYOUT", "layout" },
        { C_ENUM (SYSPROF_CALLGRAPH_CATEGORY_LOCKING), "SYSPROF_CALLGRAPH_CATEGORY_LOCKING", "locking" },
        { C_ENUM (SYSPROF_CALLGRAPH_CATEGORY_MAIN_LOOP), "SYSPROF_CALLGRAPH_CATEGORY_MAIN_LOOP", "main-loop" },
        { C_ENUM (SYSPROF_CALLGRAPH_CATEGORY_MEMORY), "SYSPROF_CALLGRAPH_CATEGORY_MEMORY", "memory" },
        { C_ENUM (SYSPROF_CALLGRAPH_CATEGORY_PAINT), "SYSPROF_CALLGRAPH_CATEGORY_PAINT", "paint" },
        { C_ENUM (SYSPROF_CALLGRAPH_CATEGORY_TYPE_SYSTEM), "SYSPROF_CALLGRAPH_CATEGORY_TYPE_SYSTEM", "type-system" },
        { C_ENUM (SYSPROF_CALLGRAPH_CATEGORY_UNWINDABLE), "SYSPROF_CALLGRAPH_CATEGORY_UNWINDABLE", "unwindable" },
        { C_ENUM (SYSPROF_CALLGRAPH_CATEGORY_WINDOWING), "SYSPROF_CALLGRAPH_CATEGORY_WINDOWING", "windowing" },
        { C_ENUM (SYSPROF_CALLGRAPH_CATEGORY_LAST), "SYSPROF_CALLGRAPH_CATEGORY_LAST", "last" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("SysprofCallgraphCategory"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

GType
sysprof_callgraph_flags_get_type (void)
{
    static gsize gtype_id = 0;
    static const GFlagsValue values[] = {
        { C_FLAGS (SYSPROF_CALLGRAPH_FLAGS_NONE), "SYSPROF_CALLGRAPH_FLAGS_NONE", "none" },
        { C_FLAGS (SYSPROF_CALLGRAPH_FLAGS_INCLUDE_THREADS), "SYSPROF_CALLGRAPH_FLAGS_INCLUDE_THREADS", "include-threads" },
        { C_FLAGS (SYSPROF_CALLGRAPH_FLAGS_HIDE_SYSTEM_LIBRARIES), "SYSPROF_CALLGRAPH_FLAGS_HIDE_SYSTEM_LIBRARIES", "hide-system-libraries" },
        { C_FLAGS (SYSPROF_CALLGRAPH_FLAGS_BOTTOM_UP), "SYSPROF_CALLGRAPH_FLAGS_BOTTOM_UP", "bottom-up" },
        { C_FLAGS (SYSPROF_CALLGRAPH_FLAGS_CATEGORIZE_FRAMES), "SYSPROF_CALLGRAPH_FLAGS_CATEGORIZE_FRAMES", "categorize-frames" },
        { C_FLAGS (SYSPROF_CALLGRAPH_FLAGS_IGNORE_PROCESS_0), "SYSPROF_CALLGRAPH_FLAGS_IGNORE_PROCESS_0", "ignore-process-0" },
        { C_FLAGS (SYSPROF_CALLGRAPH_FLAGS_LEFT_HEAVY), "SYSPROF_CALLGRAPH_FLAGS_LEFT_HEAVY", "left-heavy" },
        { C_FLAGS (SYSPROF_CALLGRAPH_FLAGS_MERGE_SIMILAR_PROCESSES), "SYSPROF_CALLGRAPH_FLAGS_MERGE_SIMILAR_PROCESSES", "merge-similar-processes" },
        { C_FLAGS (SYSPROF_CALLGRAPH_FLAGS_IGNORE_KERNEL_PROCESSES), "SYSPROF_CALLGRAPH_FLAGS_IGNORE_KERNEL_PROCESSES", "ignore-kernel-processes" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_flags_register_static (g_intern_static_string ("SysprofCallgraphFlags"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

/* enumerations from "sysprof-recording.h" */

GType
sysprof_recording_phase_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (SYSPROF_RECORDING_PHASE_PREPARE), "SYSPROF_RECORDING_PHASE_PREPARE", "prepare" },
        { C_ENUM (SYSPROF_RECORDING_PHASE_RECORD), "SYSPROF_RECORDING_PHASE_RECORD", "record" },
        { C_ENUM (SYSPROF_RECORDING_PHASE_AUGMENT), "SYSPROF_RECORDING_PHASE_AUGMENT", "augment" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("SysprofRecordingPhase"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

/* Generated data ends here */

