#ifndef SYSPROF_CREDITS_H
#define SYSPROF_CREDITS_H

#include <glib.h>

G_BEGIN_DECLS

static const gchar *sysprof_authors[] = {
  "Bastien Nocera",
  "Behdad Esfahbod",
  "Christian Hergert",
  "Damien Lespiau",
  "Geoffrey Keating",
  "Javier Jardón",
  "Jeffrey Stedfast",
  "Kristian Høgsberg",
  "Kristian Rietvel",
  "M Joonas Pihlaja",
  "Matthias Clasen",
  "Michel Dänzer",
  "Nix",
  "Owen Taylor",
  "Paolo Borelli",
  "Pascal Terjan",
  "Pauli Nieminen",
  "Robert Love",
  "Søren Sandmann Pedersen",
  NULL
};

static const gchar *sysprof_artists[] = {
  "Jakub Steiner",
  NULL
};

G_END_DECLS

#endif /* SYSPROF_CREDITS_H */
