/*
 * This file is generated by gdbus-codegen, do not modify it.
 *
 * The license of this code is the same as for the D-Bus interface description
 * it was derived from. Note that it links to GLib, so must comply with the
 * LGPL linking clauses.
 */

#ifndef __IPC_UNWINDER_H__
#define __IPC_UNWINDER_H__

#include <gio/gio.h>

G_BEGIN_DECLS


/* ------------------------------------------------------------------------ */
/* Declarations for org.gnome.Sysprof3.Unwinder */

#define IPC_TYPE_UNWINDER (ipc_unwinder_get_type ())
#define IPC_UNWINDER(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), IPC_TYPE_UNWINDER, IpcUnwinder))
#define IPC_IS_UNWINDER(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), IPC_TYPE_UNWINDER))
#define IPC_UNWINDER_GET_IFACE(o) (G_TYPE_INSTANCE_GET_INTERFACE ((o), IPC_TYPE_UNWINDER, IpcUnwinderIface))

struct _IpcUnwinder;
typedef struct _IpcUnwinder IpcUnwinder;
typedef struct _IpcUnwinderIface IpcUnwinderIface;

struct _IpcUnwinderIface
{
  GTypeInterface parent_iface;

  gboolean (*handle_unwind) (
    IpcUnwinder *object,
    GDBusMethodInvocation *invocation,
    GUnixFDList *fd_list,
    guint arg_stack_size,
    GVariant *arg_perf_fds,
    GVariant *arg_event_fd,
    GVariant *arg_capture_fd);

};

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (IpcUnwinder, g_object_unref)
#endif

GType ipc_unwinder_get_type (void);

GDBusInterfaceInfo *ipc_unwinder_interface_info (void);
guint ipc_unwinder_override_properties (GObjectClass *klass, guint property_id_begin);


/* D-Bus method call completion functions: */
void ipc_unwinder_complete_unwind (
    IpcUnwinder *object,
    GDBusMethodInvocation *invocation,
    GUnixFDList *fd_list);



/* D-Bus method calls: */
void ipc_unwinder_call_unwind (
    IpcUnwinder *proxy,
    guint arg_stack_size,
    GVariant *arg_perf_fds,
    GVariant *arg_event_fd,
    GVariant *arg_capture_fd,
    GUnixFDList *fd_list,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean ipc_unwinder_call_unwind_finish (
    IpcUnwinder *proxy,
    GUnixFDList **out_fd_list,
    GAsyncResult *res,
    GError **error);

gboolean ipc_unwinder_call_unwind_sync (
    IpcUnwinder *proxy,
    guint arg_stack_size,
    GVariant *arg_perf_fds,
    GVariant *arg_event_fd,
    GVariant *arg_capture_fd,
    GUnixFDList  *fd_list,
    GUnixFDList **out_fd_list,
    GCancellable *cancellable,
    GError **error);



/* ---- */

#define IPC_TYPE_UNWINDER_PROXY (ipc_unwinder_proxy_get_type ())
#define IPC_UNWINDER_PROXY(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), IPC_TYPE_UNWINDER_PROXY, IpcUnwinderProxy))
#define IPC_UNWINDER_PROXY_CLASS(k) (G_TYPE_CHECK_CLASS_CAST ((k), IPC_TYPE_UNWINDER_PROXY, IpcUnwinderProxyClass))
#define IPC_UNWINDER_PROXY_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), IPC_TYPE_UNWINDER_PROXY, IpcUnwinderProxyClass))
#define IPC_IS_UNWINDER_PROXY(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), IPC_TYPE_UNWINDER_PROXY))
#define IPC_IS_UNWINDER_PROXY_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), IPC_TYPE_UNWINDER_PROXY))

typedef struct _IpcUnwinderProxy IpcUnwinderProxy;
typedef struct _IpcUnwinderProxyClass IpcUnwinderProxyClass;
typedef struct _IpcUnwinderProxyPrivate IpcUnwinderProxyPrivate;

struct _IpcUnwinderProxy
{
  /*< private >*/
  GDBusProxy parent_instance;
  IpcUnwinderProxyPrivate *priv;
};

struct _IpcUnwinderProxyClass
{
  GDBusProxyClass parent_class;
};

GType ipc_unwinder_proxy_get_type (void);

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (IpcUnwinderProxy, g_object_unref)
#endif

void ipc_unwinder_proxy_new (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data);
IpcUnwinder *ipc_unwinder_proxy_new_finish (
    GAsyncResult        *res,
    GError             **error);
IpcUnwinder *ipc_unwinder_proxy_new_sync (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error);

void ipc_unwinder_proxy_new_for_bus (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data);
IpcUnwinder *ipc_unwinder_proxy_new_for_bus_finish (
    GAsyncResult        *res,
    GError             **error);
IpcUnwinder *ipc_unwinder_proxy_new_for_bus_sync (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error);


/* ---- */

#define IPC_TYPE_UNWINDER_SKELETON (ipc_unwinder_skeleton_get_type ())
#define IPC_UNWINDER_SKELETON(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), IPC_TYPE_UNWINDER_SKELETON, IpcUnwinderSkeleton))
#define IPC_UNWINDER_SKELETON_CLASS(k) (G_TYPE_CHECK_CLASS_CAST ((k), IPC_TYPE_UNWINDER_SKELETON, IpcUnwinderSkeletonClass))
#define IPC_UNWINDER_SKELETON_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), IPC_TYPE_UNWINDER_SKELETON, IpcUnwinderSkeletonClass))
#define IPC_IS_UNWINDER_SKELETON(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), IPC_TYPE_UNWINDER_SKELETON))
#define IPC_IS_UNWINDER_SKELETON_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), IPC_TYPE_UNWINDER_SKELETON))

typedef struct _IpcUnwinderSkeleton IpcUnwinderSkeleton;
typedef struct _IpcUnwinderSkeletonClass IpcUnwinderSkeletonClass;
typedef struct _IpcUnwinderSkeletonPrivate IpcUnwinderSkeletonPrivate;

struct _IpcUnwinderSkeleton
{
  /*< private >*/
  GDBusInterfaceSkeleton parent_instance;
  IpcUnwinderSkeletonPrivate *priv;
};

struct _IpcUnwinderSkeletonClass
{
  GDBusInterfaceSkeletonClass parent_class;
};

GType ipc_unwinder_skeleton_get_type (void);

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (IpcUnwinderSkeleton, g_object_unref)
#endif

IpcUnwinder *ipc_unwinder_skeleton_new (void);


G_END_DECLS

#endif /* __IPC_UNWINDER_H__ */
