/* sysprof-window.h
 *
 * Copyright 2023 Christian Hergert <chergert@redhat.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * SPDX-License-Identifier: GPL-3.0-or-later
 */

#pragma once

#include <adwaita.h>

#include <sysprof.h>

#include "sysprof-application.h"
#include "sysprof-recording-template.h"
#include "sysprof-session.h"

G_BEGIN_DECLS

#define SYSPROF_TYPE_WINDOW (sysprof_window_get_type())

G_DECLARE_FINAL_TYPE (SysprofWindow, sysprof_window, SYSPROF, WINDOW, AdwApplicationWindow)

GtkWidget       *sysprof_window_new          (SysprofApplication       *app,
                                              SysprofDocument          *document);
void             sysprof_window_open_file    (GtkWindow                *transient_for,
                                              SysprofRecordingTemplate *template);
void             sysprof_window_open         (SysprofApplication       *app,
                                              SysprofRecordingTemplate *template,
                                              GFile                    *file);
void             sysprof_window_open_fd      (SysprofApplication       *app,
                                              SysprofRecordingTemplate *template,
                                              int                       fd);
SysprofDocument *sysprof_window_get_document (SysprofWindow            *self);
SysprofSession  *sysprof_window_get_session  (SysprofWindow            *self);

G_END_DECLS
